///////////////////////////////////////////////////////////////////////////////
// Name:        animate.cpp
// Purpose:     Implementation of wxAnimation classes
// Author:      Julian Smart and Guillermo Rodriguez Garcia
// Modified by:
// Created:     13/8/99
// RCS-ID:      $Id$
// Copyright:   (c) Julian Smart and Guillermo Rodriguez Garcia
// Licence:     wxWindows licence
///////////////////////////////////////////////////////////////////////////////

#ifdef __GNUG__
    #pragma implementation "animate.h"
#endif

#include "wx/wxprec.h"

#ifdef    __BORLANDC__
  #pragma hdrstop
#endif  //__BORLANDC__

#include <wx/image.h>

#include <wx/animate.h>
#include <wx/gifdecod.h>
#include <wx/wfstream.h>

/*
 * wxAnimationPlayer
 */

IMPLEMENT_CLASS(wxAnimationPlayer, wxObject)

wxAnimationPlayer::wxAnimationPlayer(wxAnimationBase *animation, bool destroyAnimation)
{
    m_animation = animation;
    m_destroyAnimation = destroyAnimation;
    m_currentFrame = 0;
    m_window = (wxWindow*) NULL;
    m_position = wxPoint(0, 0);
    m_looped = TRUE;
    m_isPlaying = TRUE;
    m_timer.SetPlayer(this);
}

wxAnimationPlayer::~wxAnimationPlayer()
{
    Stop();

    if (m_destroyAnimation)
        delete m_animation;
}

void wxAnimationPlayer::SetAnimation(wxAnimationBase* animation, bool destroyAnimation)
{
    ClearCache();
    if (m_destroyAnimation)
        delete m_animation;
    m_animation = animation;
    m_destroyAnimation = destroyAnimation;
}

// Play
bool wxAnimationPlayer::Play(wxWindow& window, const wxPoint& pos, bool looped)
{
    m_window = & window;

    if (!m_animation || !m_animation->IsValid())
        return FALSE;

    if (m_frames.Number() == 0)
    {
        if (!Build())
        {
            wxLogWarning("wxAnimationPlayer::Play: could not build the image cache.");
            return FALSE;
        }
    }
    m_currentFrame = 0;

    PlayFrame();

    return TRUE;
}

// Build animation (list of wxImages). If not called before Play
// is called, Play will call this automatically.
bool wxAnimationPlayer::Build()
{
    ClearCache();
    if (m_animation)
    {
        int n = GetFrameCount();
        int i;
        for (i = 0; i < n; i++)
        {
            wxBitmap* bitmap = NULL;
            wxImage* image = GetFrame(i);
            if (image)
            {
                bitmap = new wxBitmap(image->ConvertToBitmap());
                delete image;
                if (bitmap)
                    m_frames.Append(bitmap);
                else
                    return FALSE;
            }
            else
                return FALSE;
        }
        return TRUE;
    }
    else
        return FALSE;
}

// Stop the animation
void wxAnimationPlayer::Stop()
{
    m_timer.Stop();
    m_isPlaying = FALSE;
}

int wxAnimationPlayer::GetFrameCount() const
{
    if (m_animation)
        return m_animation->GetFrameCount();
    else
        return 0;
}

wxImage* wxAnimationPlayer::GetFrame(int i) const
{
    if (m_animation)
        return m_animation->GetFrame(i);
    else
        return (wxImage*) NULL;
}

wxAnimationDisposal wxAnimationPlayer::GetDisposalMethod(int i) const
{
    if (m_animation)
        return m_animation->GetDisposalMethod(i);
    else
        return wxANIM_UNSPECIFIED;
}

wxRect wxAnimationPlayer::GetFrameRect(int i) const
{
    if (m_animation)
        return m_animation->GetFrameRect(i);
    else
        return wxRect(0, 0, 0, 0);
}

int wxAnimationPlayer::GetDelay(int i) const
{
    if (m_animation)
        return m_animation->GetDelay(i);
    else
        return 0;
}

wxSize wxAnimationPlayer::GetLogicalScreenSize() const
{
    if (m_animation)
        return m_animation->GetLogicalScreenSize();
    else
        return wxSize(0, 0);
}

wxColour wxAnimationPlayer::GetBackgroundColour() const
{
    if (m_animation)
        return m_animation->GetBackgroundColour();
    else
        return wxColour();
}

// Play the frame
bool wxAnimationPlayer::PlayFrame(int frame, wxWindow& window, wxPoint& pos)
{
    wxClientDC dc(& window);

    // Draw the background colour
    DrawBackground(dc, pos, GetBackgroundColour());

    // Draw all intermediate frames that haven't been removed from the
    // animation
    int i;
    for (i = 0; i < (frame - 1); i++)
    {
        if ((GetDisposalMethod(i) == wxANIM_DONOTREMOVE) || (GetDisposalMethod(i) == wxANIM_UNSPECIFIED))
        {
            DrawFrame(i, dc, pos);
        }
    }
    DrawFrame(frame, dc, pos);

    return TRUE;
}

bool wxAnimationPlayer::PlayFrame()
{
    m_isPlaying = TRUE;

    PlayFrame(GetCurrentFrame(), * GetWindow(), GetPosition());

    // Set the timer for the next frame
    m_timer.Start(GetDelay(GetCurrentFrame()));

    m_currentFrame ++;

    if (m_currentFrame == GetFrameCount())
    {
        // Should a non-looped animation display the last frame?
        if (!m_looped)
        {
            m_timer.Stop();
            m_isPlaying = FALSE;
        }
        else
            m_currentFrame = 0;
    }

    return TRUE;
}

// Clear the wxImage cache
void wxAnimationPlayer::ClearCache()
{
    wxNode* node = m_frames.First();
    while (node)
    {
        wxNode* next = node->Next();
        wxBitmap* bitmap = (wxBitmap*) node->Data();
        delete bitmap;
        delete node;
        node = next;
    }
}

// Draw the background colour
void wxAnimationPlayer::DrawBackground(wxDC& dc, const wxPoint& pos, const wxColour& colour)
{
    wxASSERT_MSG( (m_animation != NULL), "Animation not present in wxAnimationPlayer");
    wxASSERT_MSG( (m_frames.Number() != 0), "Animation cache not present in wxAnimationPlayer");

    // Optimization: if the first frame fills the whole area, and is non-transparent,
    // don't bother drawing the background

    wxBitmap* firstBitmap = (wxBitmap*) m_frames.First()->Data() ;
    wxSize screenSize = GetLogicalScreenSize();
    if (!firstBitmap->GetMask() && (firstBitmap->GetWidth() == screenSize.x) && (firstBitmap->GetHeight() == screenSize.y))
    {
        return;
    }

    wxBrush brush(colour, wxSOLID);
    wxPen pen(colour, wxSOLID, 1);
    dc.SetBrush(brush);
    dc.SetPen(pen);
    dc.SetLogicalFunction(wxCOPY);

    dc.DrawRectangle(pos.x, pos.y, screenSize.x, screenSize.y);
}

// Draw this frame
void wxAnimationPlayer::DrawFrame(int frame, wxDC& dc, const wxPoint& pos)
{
    wxASSERT_MSG( (m_animation != NULL), "Animation not present in wxAnimationPlayer");
    wxASSERT_MSG( (m_frames.Number() != 0), "Animation cache not present in wxAnimationPlayer");
    wxASSERT_MSG( (m_frames.Nth(frame) != (wxNode*) NULL), "Image not present in wxAnimationPlayer::DrawFrame");

    wxBitmap* bitmap = (wxBitmap*) m_frames.Nth(frame)->Data() ;

    wxRect rect = GetFrameRect(frame);

    dc.DrawBitmap(* bitmap, pos.x + rect.x, pos.y + rect.y, (bitmap->GetMask() != NULL));
}

void wxAnimationTimer::Notify()
{
    m_player->PlayFrame();
}

/*
 * wxAnimationBase
 */

IMPLEMENT_ABSTRACT_CLASS(wxAnimationBase, wxObject)

/*
 * wxGIFAnimation
 */

IMPLEMENT_CLASS(wxGIFAnimation, wxAnimationBase)

wxGIFAnimation::wxGIFAnimation()
{
    m_decoder = (wxGIFDecoder*) NULL;
}

wxGIFAnimation::~wxGIFAnimation()
{
    delete m_decoder;
}

int wxGIFAnimation::GetFrameCount() const
{
    wxASSERT_MSG( (m_decoder != (wxGIFDecoder*) NULL), "m_decoder must be non-NULL");

    return m_decoder->GetNumberOfFrames();
}

wxImage* wxGIFAnimation::GetFrame(int i) const
{
    wxASSERT_MSG( (m_decoder != (wxGIFDecoder*) NULL), "m_decoder must be non-NULL");

    m_decoder->GoFrame(i);

    wxImage* image = new wxImage;
    m_decoder->ConvertToImage(image);
    return image;
}

wxAnimationDisposal wxGIFAnimation::GetDisposalMethod(int i) const
{
    wxASSERT_MSG( (m_decoder != (wxGIFDecoder*) NULL), "m_decoder must be non-NULL");

    m_decoder->GoFrame(i);

    int disposalMethod = m_decoder->GetDisposalMethod();
    return (wxAnimationDisposal) disposalMethod;
}

wxRect wxGIFAnimation::GetFrameRect(int i) const
{
    wxASSERT_MSG( (m_decoder != (wxGIFDecoder*) NULL), "m_decoder must be non-NULL");

    m_decoder->GoFrame(i);

    wxRect rect(m_decoder->GetLeft(), m_decoder->GetTop(), m_decoder->GetWidth(), m_decoder->GetHeight());
    return rect;
}

int wxGIFAnimation::GetDelay(int i) const
{
    wxASSERT_MSG( (m_decoder != (wxGIFDecoder*) NULL), "m_decoder must be non-NULL");

    m_decoder->GoFrame(i);
    return m_decoder->GetDelay();
}

wxSize wxGIFAnimation::GetLogicalScreenSize() const
{
    wxASSERT_MSG( (m_decoder != (wxGIFDecoder*) NULL), "m_decoder must be non-NULL");

    return wxSize(m_decoder->GetLogicalScreenWidth(), m_decoder->GetLogicalScreenHeight());
}

wxColour wxGIFAnimation::GetBackgroundColour() const
{
    wxASSERT_MSG( (m_decoder != (wxGIFDecoder*) NULL), "m_decoder must be non-NULL");

    // TODO: how to get the colour from the index?
    return wxColour(0, 0, 0);
}

bool wxGIFAnimation::IsValid() const
{
    return ((m_decoder != NULL) && (m_decoder->IsAnimation()));
}

bool wxGIFAnimation::LoadFile(const wxString& filename)
{
    if (m_decoder)
        delete m_decoder;
    m_decoder = NULL;

    if (wxFileExists(filename))
    {
        wxFileInputStream stream(filename);
        m_decoder = new wxGIFDecoder(& stream, TRUE);

        if (m_decoder->ReadGIF() != E_OK)
        {
            delete m_decoder;
            m_decoder = NULL;
            return FALSE;
        }

        if (!m_decoder->IsAnimation())
        {
            delete m_decoder;
            m_decoder = NULL;

            return FALSE;
        }
        else
            return TRUE;
    }
    else
        return FALSE;
}
